/*
 * steve diverdi
 * cs125 project 4
 *
 * tftp common functions
 */


// INCLUDES ------------------------------------------------------------

#include "common.hh"

// ---------------------------------------------------------------------


/*
 * Sends an error packet and prints the error to stderr.  takes as 
 * arguments the tftp error code, a corresponding errormsg (from RFC 1350),
 * a socket id, a connection address, and the address length.
 */
void error (
    int			errorcode,
    char*		errormsg,
    int			s,
    struct sockaddr*	conn_addr,
    int			conn_len)
{
    // print the error to stderr
    cerr << endl << errormsg << " Error " << errno << endl;

    // make the error packet
    struct error_packet p;
    int len;
    makeERROR(&p, &len, errorcode, errormsg);

    // send it on its way
    sendto(s, &p, len, 0, conn_addr, conn_len);
}


// ---------------------------------------------------------------------

/*
 * Put a buffer in network byte order.
 */
void networkify (
    void*		p,
    int			len)
{
    // apply htons to each byte in the buffer
    for (int i = 0; i < len; i++)
        ((char*) p)[i] = htons(((char*) p)[i]);
}


/*
 * put a buffer in host byte order
 */
void denetworkify (
    void*		p,
    int			len)
{
    // apply ntohs to each byte in the buffer
    for (int i = 0; i < len; i++)
        ((char*) p)[i] = ntohs(((char*) p)[i]);
}


// ---------------------------------------------------------------------


/*
 * Make a read request packet.  Takes as arguments the packet to initialize,
 * the length of the packet, the name of the file to request, and the 
 * mode to use.
 */
void makeRRQ (
    struct packet*	p,
    int*		len,
    char*		filename,
    char*		mode)
{
    // init the packet to zeroes
    bzero(p, sizeof(packet));

    // put the values into the packet
    p->opcode = (uint16_t) RRQ;
    strncpy(p->data, filename, (strlen(filename) + 1));
    strncat(p->data + (strlen(filename) + 1), mode, (strlen(mode) + 1));

    // set the length
    *len = sizeof(p->opcode) + strlen(filename) + 1 + strlen(mode) + 1;

    // put in network byte order
    networkify((char*) p, *len);
}


/*
 * Make a write request packet.  Takes as arguments the packet to
 * initialize, the length of the packet, the name of the file to request,
 * and the mode to use.
 */
void makeWRQ (
    struct packet*	p,
    int*		len,
    char*		filename,
    char*		mode)
{
    // init the packet to zeroes
    bzero(p, sizeof(packet));

    // put the values into the packet
    p->opcode = (uint16_t) WRQ;
    strncpy(p->data, filename, (strlen(filename) + 1));
    strncpy((p->data + strlen(filename) + 1), mode, (strlen(mode) + 1));

    // set the length
    *len = sizeof(p->opcode) + strlen(filename) + 1 + strlen(mode) + 1;

    // put in network byte order
    networkify((char*) p, *len);
}


/*
 * Make a data packet.  Takes as arguments the packet to initialize, the
 * length of the packet, the block number, the data to put in the packet,
 * and the length of the data.
 */
void makeDATA (
    struct data_packet*	p,
    int*		len,
    int			block,
    char*		data, 
    int			dlen)
{
    // init the packet to zeroes
    bzero(p, sizeof(data_packet));

    // put the values into the packet
    p->opcode = (uint16_t) DATA;
    p->block  = (uint16_t) block;
    strncpy(p->data, data, dlen);

    // set the length
    *len = sizeof(p->opcode) + sizeof(p->block) + dlen;

    // put in network byte order
    networkify((char*) p, *len);
}


/*
 * Make an acknowledgement packet.  Takes as arguments the packet to
 * initialize, the length of the packet, and the block number.
 */
void makeACK (
    struct ack_packet*	p,
    int*		len,
    int			block)
{
    // init the packet to zeroes
    bzero(p, sizeof(ack_packet));

    // put the value into the packet
    p->opcode = (uint16_t) ACK;
    p->block  = (uint16_t) block;

    // set the length
    *len = sizeof(p->opcode) + sizeof(p->block);

    // put in network byte order
    networkify((char*) p, *len);
}


/*
 * Make an error packet.  Takes as arguments the packet to initialize,
 * the length of the packet, the error code, error message and the length
 * of the message.
 */
void makeERROR (
    struct error_packet*	p,
    int*		len,
    int			errorcode,
    char*		errormsg)
{
    // init the packet to zeroes
    bzero(p, sizeof(error_packet));

    // put the value into the packet
    p->opcode = (uint16_t) ERROR;
    p->errorcode = (uint16_t) errorcode;
    strncpy(p->errmsg, errormsg, (strlen(errormsg) + 1));

    // set the length
    *len = sizeof(p->opcode) + sizeof(p->errorcode) + strlen(errormsg) + 1;

    // put in network byte order
    networkify((char*) p, *len);
}


// ---------------------------------------------------------------------


/*
 * return the packet's opcode field
 */
int getType (
    struct packet*	packet)
{
    return (int) packet->opcode;
}


/*
 * returns the packet's block field, if it's an acknowledgement or data
 * packet.  returns -1 otherwise.
 */
int getBlock (
    struct packet*	packet)
{
    if (getType(packet) == ACK)
        return (int) ((struct ack_packet*) packet)->block;

    else if (getType(packet) == DATA)
        return (int) ((struct data_packet*) packet)->block;

    else
        return -1;
}


/*
 * returns the packet's filename, if it's a read request or write request
 * packet.  returns NULL otherwise
 */
char* getFilename (
    struct packet*	packet)
{
    if ((getType(packet) == RRQ) ||
        (getType(packet) == WRQ))
    {
        int len = strlen(packet->data) + 1;
        char* buffer = new char[len];
        strncpy(buffer, packet->data, len);

        return buffer;
    }
    else
        return NULL;
}


/*
 * returns the packet's mode, if it's a read request or write request
 * packet.  returns -1 otherwise.
 */
int getMode (
    struct packet*	packet)
{
    if ((getType(packet) == RRQ) ||
        (getType(packet) == WRQ))
    {
        int offset = strlen(packet->data) + 1;
        int len = strlen(packet->data + offset) + 1;
        char* buffer = new char[len];
        strncpy(buffer, (packet->data + offset), len);

        int result;
        if (strncmp(buffer, "octet", 6) == 0)
            result = OCTET;
        else if (strncmp(buffer, "netascii", 9) == 0)
            result = NETASCII;
        else if (strncmp(buffer, "mail", 5) == 0)
            result = MAIL;
        else
            result = -1;

        delete[] buffer;
        return result;
    }
    else
        return -1;
}


/*
 * returns the packet's error message, if it's an error packet.  returns
 * NULL otherwise.
 */
char* getErrormsg (
    struct packet*	packet)
{
    if (getType(packet) == ERROR)
    {
//        int len = strlen(((struct error_packet*) packet)->errmsg) + 1;
//        char* buffer = new char[len];
//        strncpy(buffer, ((struct error_packet*) packet)->errmsg, len);

        char* buffer = ((struct error_packet*) packet)->errmsg;

        return buffer;
    }
    else
        return NULL;
}


// ---------------------------------------------------------------------


/*
 * Checks a socket to see if it has a packet waiting to be received
 * within the time out value.  Returns false if the time out is reached
 * before incoming traffic is found, true otherwise.
 */
bool timeout (
    int                 s)
{
    // make an fd_set with our socket marked
    fd_set set;
    FD_ZERO(&set);
    FD_SET(s, &set);

    // check our socket within the time out
    if (select((s + 1), &set, NULL, NULL, &TIMEOUT) == -1)
    {
        cerr << "timeout: select " << errno << endl;
    }

    // returns if our socket is still set in fd_set
    return (! FD_ISSET(s, &set));
}


// ---------------------------------------------------------------------


/*
 * Forms and sends an acknowledgement packet to the connection.  Takes
 * as input a socket id, a block number, a connection address, the
 * address length, and a bool that indicates if a response is expected
 * or not.  Returns 0 if successful, -1 otherwise.
 */
int sendACK (
    int                 s,
    int                 block,
    struct sockaddr*    conn_addr,
    int                 conn_len, 
    bool                expectResponse)   
{
    // make the acknowledgement packet
    struct ack_packet outgoing;
    int len;
    makeACK(&outgoing, &len, block);

    // send the packet to the connection
    if (sendto(s, &outgoing, len, 0, conn_addr, conn_len) < 0)
    {
        error(TNOTDEF, "Couldn't send packet", s, conn_addr, conn_len);
        return -1;
    }
     
    // if a reponse is expected, and the timeout is reached before a
    // response is availabe, resend the packet
    if (expectResponse)
    {
        bool response = false;
        for (int i = 0; (i < RETRIES) && (! response); i++)
        {
            // check if a response is ready within the timeout
            response = ! timeout(s);

            if (! response)
            {
                // resend the packet
                if (sendto(s, &outgoing, len, 0, conn_addr, conn_len) < 0)
                {
                    error(TNOTDEF, "Couldn't send packet", s, conn_addr, 
                          conn_len);
                    return -1;
                }
            }
        }
         
        // if the timeout is reached again, error and fail
        if (! response)
        {
            error(TNOTDEF, "Connection timed out", s, conn_addr, conn_len);
            return -1;
        }
    }
     
    return 0;
}


/*
 * Receive an acknowledgement packet from a connection.  Takes as
 * arguments a socket id, an expected block number, the expected 
 * port number, a connection address, and the address length.  returns
 * 0 on success, -1 on failure.
 */
int recvACK (
    int			s,
    int			blockNum,
    int			conn_tid,
    struct sockaddr*	conn_addr,
    int*		conn_len)
{
    // make room for the packet
    struct ack_packet incoming;
    int ilen;

    // receive the next packet from the connection
    if ((ilen = recvfrom(s, &incoming, sizeof(incoming), 0,
                         conn_addr, conn_len)) == -1)
    {
        error(TNOTDEF, "Couldn't receive ack", s, conn_addr, *conn_len);
        return -1;
    }

    // host byte order
    denetworkify(&incoming, ilen);

    // if the port wasn't the expected port...
    if ((conn_tid != -1) &&
        (conn_tid != ((struct sockaddr_in*) conn_addr)->sin_port))
    {
        // send an error to the wrong port
        error(TBADTID, "Unkown transfer ID", s, conn_addr, *conn_len);
            
        // keep waiting on our connection for more traffic
        if (timeout(s))
        {
            error(TNOTDEF, "Connection timed out", s, conn_addr, *conn_len);
            return -1;
        }
        
        // when more traffic arrives, go get it.
        return recvACK(s, blockNum, conn_tid, conn_addr, conn_len);
    }

    // if it's an error packet bail out
    if (getType((struct packet*) &incoming) == ERROR)
    {
        cerr << "Error: " << getErrormsg((struct packet*) &incoming) << endl;
        return -1;
    }

    // if the packet isn't an acknowledgement packet or isn't the write
    // block number, bail out.
    if ((getType((struct packet*) &incoming) != ACK) ||
        (getBlock((struct packet*) &incoming) != blockNum))
    {
        error(TILGLOP, "Illegal TFTP operation", s, conn_addr, *conn_len);
        return -1;
    }

    return 0;
}


// ---------------------------------------------------------------------


/*
 * Make and send a data packet accross a connection.  Takes as arguments
 * a socket id, a block number, the block data, the length of the data,
 * the conenction address, and the length of the address.  Returns 0
 * on success and -1 on failure.
 */
int sendDATA (
    int                 s,
    int                 blockNum,
    char*               blockData,
    int                 blen,
    struct sockaddr*    conn_addr,
    int                 conn_len)
{
    // make the data packet     
    struct data_packet outgoing;
    int olen;
    makeDATA(&outgoing, &olen, blockNum, blockData, blen);

    // send the data packet along
    if (sendto(s, &outgoing, olen, 0, conn_addr, conn_len) < 0)
    {
        error(TNOTDEF, "Couldn't send packet", s, conn_addr, conn_len);
        return -1;
    }   

    bool response = false;
    for (int i = 0; (i < RETRIES) && (! response); i++)
    {
        // we expect a response to a data packet, so check if there's an
        // incoming packet within the timeout on our socket        
        response = ! timeout(s);

        if (! response)
        {
            // there wasn't, so resend the packet
            if (sendto(s, &outgoing, olen, 0, conn_addr, conn_len) < 0)
            {
                error(TNOTDEF, "Couldn't send packet", s, conn_addr, 
                      conn_len);
                return -1;
            }
        }
    }

    // if new traffic didn't arrive the last time, fail
    if (! response)
    {
        error(TNOTDEF, "Connection timed out", s, conn_addr, conn_len);
        return -1;
    }

    return 0;
}


/* 
 * Receive a data packet from our connection.  Takes as arguments a 
 * socket id, space for the incoming data packet, the length of the data
 * packet, the block number expected, the port number expected, the
 * connection address, and the length of the address.  returns 0 on
 * success and -1 on failure
 */
int recvDATA (
    int                 s,
    struct data_packet* incoming,
    int*                ilen,
    int                 blockNum,
    int                 conn_tid,
    struct sockaddr*    conn_addr,
    int*                conn_len)
{
    // receive the next packet on our socket
    if ((*ilen = recvfrom(s, incoming, sizeof(struct data_packet), 0,
                          conn_addr, conn_len)) == -1)
    {
        error(TNOTDEF, "Couldn't receive data", s, conn_addr, *conn_len);
        return -1;
    }

    // host byte order
    denetworkify(incoming, *ilen);

    // check if the expected port is the same as the packet's port
    if ((conn_tid != -1) &&
        (conn_tid != ((struct sockaddr_in*) conn_addr)->sin_port))
    {
        // send an error to the incorrect port
        error(TBADTID, "Unkown transfer ID", s, conn_addr, *conn_len);
            
        // keep waiting for more traffic on our socket
        if (timeout(s))
        {
            error(TNOTDEF, "Connection timed out", s, conn_addr, *conn_len);
            return -1;
        }
        
        // when more traffic arrives, go get it
        return recvDATA(s, incoming, ilen, blockNum, conn_tid,
                        conn_addr, conn_len);
    }   

    // if the type of the packet is an error, bail out.
    if (getType((struct packet*) incoming) == ERROR)
    {
        cerr << "Error: " << getErrormsg((struct packet*) incoming) << endl;
        return -1;
    }

    // check the packet is a data packet and has the right block number.
    if ((getType((struct packet*) incoming) != DATA) ||
        (getBlock((struct packet*) incoming) != blockNum))
    {
        error(TILGLOP, "Illegal TFTP operation", s, conn_addr, *conn_len);
        return -1;
    }
     
    return 0;
}
